from collections.abc import Collection
from typing import TYPE_CHECKING, Any, Literal, Optional, Union

from sym_metanet.blocks.base import ElementWithVars
from sym_metanet.blocks.origins import MeteredOnRamp
from sym_metanet.engines.core import EngineBase, get_current_engine
from sym_metanet.util.funcs import first
from sym_metanet.util.types import VarType

if TYPE_CHECKING:
    from sym_metanet.blocks.nodes import Node
    from sym_metanet.network import Network


class Link(ElementWithVars[VarType]):
    """Highway link between two nodes [1, Section 3.2.1]. Links represent stretch of
    highway with similar traffic characteristics and no road changes (e.g., same number
    of lanes and maximum speed).

    References
    ----------
    [1] Hegyi, A., 2004, "Model predictive control for integrating traffic control
        measures", Netherlands TRAIL Research School.
    """

    __slots__ = ("N", "lam", "L", "rho_max", "rho_crit", "v_free", "a", "turnrate")
    _states = {"rho", "v"}

    def __init__(
        self,
        nb_segments: int,
        lanes: Union[VarType, int],
        length: Union[VarType, float],
        maximum_density: Union[VarType, float],
        critical_density: Union[VarType, float],
        free_flow_velocity: Union[VarType, float],
        a: Union[VarType, float],
        turnrate: Union[VarType, float] = 1.0,
        name: Optional[str] = None,
    ) -> None:
        """Creates an instance of a METANET link.

        Parameters
        ----------
        nb_segments : int
            Number of segments in this highway link, i.e., `N`.
        lanes : int or variable
            Number of lanes in each segment, i.e., `lam`.
        lengths : float or variable
            Length of each segment in the link, i.e., `L`.
        maximum density : float or variable
            Maximum density that the link can withstand, i.e., `rho_max`.
        critical_densities : float or variable
            Critical density at which the traffic flow is maximal, i.e., `rho_crit`.
        free_flow_velocities : float or variable
            Average speed of cars when traffic is freely flowing, i.e., `v_free`.
        a : float or variable
            Model parameter in the computations of the equivalent speed [1, Equation
            3.4].
        turnrate : float or variable, optional
            Fraction of the total flow that enters this link via the upstream node. Only
            relevant if multiple exiting links are attached to the same node, in order
            to split the flow according to these rates. Needs not be normalized. By
            default, all links have equal rates.
        name : str, optional
            Name of this link, by default `None`.

        References
        ----------
        [1] Hegyi, A., 2004, "Model predictive control for integrating traffic control
            measures", Netherlands TRAIL Research School.
        """
        super().__init__(name)
        self.N = nb_segments
        self.lam = lanes
        self.L = length
        self.rho_max = maximum_density
        self.rho_crit = critical_density
        self.v_free = free_flow_velocity
        self.a = a
        self.turnrate = turnrate

    def init_vars(
        self,
        init_conditions: Optional[dict[str, VarType]] = None,
        engine: Optional[EngineBase] = None,
        positive_init_speed: bool = False,
        positive_init_density: bool = False,
        **_,
    ) -> None:
        """For each segment in the link, initializes
         - `rho`: densities (state)
         - `v`: speeds (state).

        Parameters
        ----------
        init_conditions : dict[str, variable], optional
            Provides name-variable tuples to initialize states, actions and disturbances
            with specific values. These values must be compatible with the symbolic
            engine in type and shape. If not provided, variables are initialized
            automatically.
        engine : EngineBase, optional
            The engine to be used. If `None`, the current engine is used.
        positive_init_speed, positive_init_density : bool, optional
            If `True`, forces the initial speed/density to be positive, e.g., as
            `v = max(0, v)`. METANET is in fact known to sometime yield negative
            quantities, which are infeasible in reality.
        """
        if init_conditions is None:
            init_conditions = {}
        if engine is None:
            engine = get_current_engine()

        self.states: dict[str, VarType] = {
            name: (
                init_conditions[name]
                if name in init_conditions
                else engine.var(f"{name}_{self.name}", self.N)
            )
            for name in ("rho", "v")
        }

        if positive_init_density:
            self.states["rho"] = engine.max(0, self.states["rho"])
        if positive_init_speed:
            self.states["v"] = engine.max(0, self.states["v"])

    def get_flow(self, engine: Optional[EngineBase] = None, **kwargs) -> VarType:
        """Gets the flow in this link's segments.

        Parameters
        ----------
        engine : EngineBase, optional
            The engine to be used. If `None`, the current engine is used.

        Returns
        -------
        variable
            The flow in this link.
        """
        if engine is None:
            engine = get_current_engine()
        return engine.links.get_flow(self.states["rho"], self.states["v"], self.lam)

    def step_dynamics(
        self,
        net: "Network",
        tau: Union[VarType, float],
        eta: Union[VarType, float],
        kappa: Union[VarType, float],
        T: Union[VarType, float],
        delta: Union[None, VarType, float] = None,
        phi: Union[None, VarType, float] = None,
        engine: Optional[EngineBase] = None,
        positive_next_speed: bool = True,
        positive_next_density: bool = False,
        **_,
    ) -> dict[str, VarType]:
        """Steps the dynamics of this link.

        Parameters
        ----------
        net : Network
            The network the link belongs to.
        tau : float or variable
            Model parameter for the speed relaxation term.
        eta : float or variable
            Model parameter for the speed anticipation term.
        kappa : float or variable
            Model parameter for the speed anticipation term.
        T : float or variable
            Sampling time.
        delta : float or variable, optional
            Model parameter for merging phenomenum. By default, not considered.
        phi : float or variable, optional
            Model parameter for lane drop phenomenum. By defaul, not considered.
        engine : EngineBase, optional
            The engine to be used. If `None`, the current engine is used.
        positive_next_speed, positive_next_density : bool, optional
            If `True`, forces the speed/density at the next time step to be positive,
            e.g., as `v+ = max(0, v+)`. METANET is in fact known to sometime yield
            negative quantities, which are infeasible in reality.

        Returns
        -------
        Dict[str, variable]
            A dict with the states of the link (speeds and densities) at the next time
            step.
        """
        if engine is None:
            engine = get_current_engine()

        node_up, node_down = net.nodes_by_link[self]  # type: ignore[index]
        rho = self.states["rho"]
        v = self.states["v"]
        q = self.get_flow(engine)

        # get upstream flow and speed, and downstream density
        v0, q0 = node_up.get_upstream_speed_and_flow(net, self, engine, T=T)
        rhoN_1 = node_down.get_downstream_density(net, engine)
        if self.N > 1:
            q_up = engine.vcat(q0, q[:-1])
            v_up = engine.vcat(v0, v[:-1])
            rho_down = engine.vcat(rho[1:], rhoN_1)
        else:
            q_up = q0
            v_up = v0
            rho_down = rhoN_1

        # check for ramp merging in this link's upstream node with other
        # entering links.
        q_ramp = None
        if (
            delta is not None
            and node_up in net.origins_by_node
            and any(net.in_links(node_up))
        ):
            origin = net.origins_by_node[node_up]  # type: ignore[index]
            if isinstance(origin, MeteredOnRamp):
                q_ramp = origin.get_flow(net, T, engine)

        # check for lane drops in the next link (only if one link downstream)
        lanes_drop = None
        if phi is not None:
            links_down: Collection[
                tuple["Node", "Node", "Link[VarType]"]
            ] = net.out_links(node_down)
            if len(links_down) == 1:
                link_down = first(links_down)[-1]
                lanes_drop = self.lam - link_down.lam  # type: ignore[operator]
            if lanes_drop == 0:
                lanes_drop = None

        # step densities
        rho_next = engine.links.step_density(rho, q, q_up, self.lam, self.L, T)

        # step speeds
        Veq = self._get_equilibrium_speed(engine, rho)
        v_next = engine.links.step_speed(
            v,
            v_up,
            rho,
            rho_down,
            Veq,
            self.lam,
            self.L,
            tau,
            eta,
            kappa,
            T,
            q_ramp,
            delta,
            lanes_drop,
            phi,
            self.rho_crit,
        )
        if positive_next_density:
            rho_next = engine.max(0, rho_next)
        if positive_next_speed:
            v_next = engine.max(0, v_next)
        return {"rho": rho_next, "v": v_next}

    def _get_equilibrium_speed(self, engine: EngineBase, rho: VarType) -> VarType:
        """Internal utility to compute the equilibrium speed for the link's segments."""
        return engine.links.Veq(rho, self.v_free, self.rho_crit, self.a)


class LinkWithVsl(Link[VarType]):
    """Highway link between two nodes and whose segments are equipped with Variable
    Speed Limit signs [1, Section 3.3.1].

    References
    ----------
    [1] Hegyi, A., 2004, "Model predictive control for integrating traffic control
        measures", Netherlands TRAIL Research School.
    """

    __slots__ = ("vsl", "alpha")
    _actions = {"v_ctrl"}

    def __init__(
        self, *args: Any, segments_with_vsl: set[int], alpha: float, **kwargs: Any
    ) -> None:
        """Creates an instance of a METANET link with VSL signs.

        Parameters
        ----------
        args, kwargs
            See base class `Link` for the other parameters.
            Average speed of cars when traffic is freely flowing, i.e., `v_free`.
        segments_with_vsl : set of ints
            The set of segment indices that are equipped with a VSL sign (0-based).
        alpha : float
            Non-compliance factor to the indicated speed limit.

        References
        ----------
        [1] Hegyi, A., 2004, "Model predictive control for integrating traffic control
            measures", Netherlands TRAIL Research School.
        """
        super().__init__(*args, **kwargs)
        self.vsl = sorted(segments_with_vsl)  # has to be a list for indexing vectors
        self.alpha = alpha
        for index in self.vsl:
            if index >= self.N or index < 0:
                raise ValueError(f"Invalid segment index {index} for VSL sign.")

    def init_vars(
        self,
        init_conditions: Optional[dict[str, VarType]] = None,
        engine: Optional[EngineBase] = None,
        **kwargs: Any,
    ) -> None:
        """For each segment in the link, initializes
         - `rho`: densities (state)
         - `v`: speeds (state)
        and, if the segment is equipped with a VSL sign, initializes also the control
        speed
         - `v_ctrl` (action).

        Parameters
        ----------
        init_conditions : dict[str, variable], optional
            Provides name-variable tuples to initialize states, actions and disturbances
            with specific values. These values must be compatible with the symbolic
            engine in type and shape. If not provided, variables are initialized
            automatically.
        engine : EngineBase, optional
            The engine to be used. If `None`, the current engine is used.
        kwargs
            See method of base class `Link`.
        """
        if init_conditions is None:
            init_conditions = {}
        if engine is None:
            engine = get_current_engine()
        super().init_vars(init_conditions, engine, **kwargs)
        self.actions: dict[str, VarType] = {
            "v_ctrl": init_conditions["v_ctrl"]
            if "v_ctrl" in init_conditions
            else engine.var(f"v_ctrl_{self.name}", len(self.vsl))
        }

    def _get_equilibrium_speed(self, engine: EngineBase, rho: VarType) -> VarType:
        return engine.links.controlled_Veq(
            rho,
            self.actions["v_ctrl"],
            self.vsl,
            self.alpha,
            self.v_free,
            self.rho_crit,
            self.a,
        )


class SimplifiedLinkWithVsl(LinkWithVsl[VarType]):
    """A simplified version of a link with VSL sign, where the equilibrium speed of the
    link is the direct control action (instead of controlling the displayed speed).

    See `LinkWithVsl` for the original version."""

    __slots__ = ("vsl", "alpha", "V_eq_type")
    _actions = {"V"}

    def __init__(
        self,
        *args: Any,
        segments_with_vsl: set[int],
        V_eq_type: Literal["limited", "unlimited"] = "limited",
        **kwargs: Any,
    ) -> None:
        """Creates an instance of a (simplified) METANET link with VSL signs.

        Parameters
        ----------
        args, kwargs
            See base class `Link` for the other parameters.
            Average speed of cars when traffic is freely flowing, i.e., `v_free`.
        segments_with_vsl : set of ints
            The set of segment indices that are equipped with a VSL sign (0-based).
        V_eq_type : "limited" or "unlimited"
            If "limited", the equilibrium speed action is capped by the standard
            equilibrium speed of the link, which is a function of the link density. If
            "unlimited", this cap is not forced, and it's up to the user to satisfy it.
            By default, "limited" is selected.

        References
        ----------
        [1] Hegyi, A., 2004, "Model predictive control for integrating traffic control
            measures", Netherlands TRAIL Research School.
        """
        super().__init__(
            *args, segments_with_vsl=segments_with_vsl, alpha=float("nan"), **kwargs
        )
        self.V_eq_type = V_eq_type

    def init_vars(
        self,
        init_conditions: Optional[dict[str, VarType]] = None,
        engine: Optional[EngineBase] = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Initializes as control action the equilibrium speed `V` of the link instead
        of the control speed `v_ctrl`."""
        if init_conditions is None:
            init_conditions = {}
        if engine is None:
            engine = get_current_engine()

        super().init_vars(init_conditions, engine, *args, **kwargs)

        del self.actions["v_ctrl"]
        self.actions["V"] = (
            init_conditions["V"]
            if "V" in init_conditions
            else engine.var(f"V_{self.name}", len(self.vsl))
        )

    def _get_equilibrium_speed(self, engine: EngineBase, rho: VarType) -> VarType:
        if engine is None:
            engine = get_current_engine()
        V_ctrl = self.actions["V"]
        Veq = engine.links.Veq(rho, self.v_free, self.rho_crit, self.a)
        if self.V_eq_type == "unlimited":
            Veq[self.vsl] = V_ctrl
        else:
            Veq[self.vsl] = engine.min(Veq[self.vsl], V_ctrl)
        return Veq



from collections.abc import Collection, Generator, Iterable
from typing import TYPE_CHECKING, Any, Optional, Union

import networkx as nx

from sym_metanet.util.types import VarType

if TYPE_CHECKING:
    from sym_metanet.blocks.links import Link
    from sym_metanet.blocks.nodes import Node


LINKENTRY = "link"
ORIGINENTRY = "origin"
DESTINATIONENTRY = "destination"


class OutLinkViewWrapper(nx.classes.reportviews.OutEdgeView):
    """Wrapper around `networkx`'s outward edge view to facilitate operations
    with link."""

    def __getitem__(self, e: tuple["Node", "Node"]) -> "Link[VarType]":
        return super().__getitem__(e)[LINKENTRY]

    def __iter__(self) -> Generator[tuple["Node", "Node", "Link[VarType]"], None, None]:
        for un, dns in self._nodes_nbrs():
            for dn, l in dns.items():
                yield (un, dn, l[LINKENTRY])

    def __call__(
        self,
        nbunch: Union[None, "Node", Iterable["Node"]] = None,
        data: Union[bool, str] = LINKENTRY,
        default: Optional[dict[str, Any]] = None,
    ) -> Collection[tuple["Node", "Node", "Link[VarType]"]]:
        return super().__call__(nbunch=nbunch, data=data, default=default)


class InLinkViewWrapper(nx.classes.reportviews.InEdgeView):
    """Wrapper around `networkx`'s inward edge view to facilitate operations
    with link."""

    def __getitem__(self, e: tuple["Node", "Node"]) -> "Link[VarType]":
        return super().__getitem__(e)[LINKENTRY]

    def __iter__(self) -> Generator[tuple["Node", "Node", "Link[VarType]"], None, None]:
        for un, dns in self._nodes_nbrs():
            for dn, l in dns.items():
                yield (un, dn, l[LINKENTRY])

    def __call__(
        self,
        nbunch: Union[None, "Node", Iterable["Node"]] = None,
        data: Union[bool, str] = LINKENTRY,
        default: Optional[dict[str, Any]] = None,
    ) -> Collection[tuple["Node", "Node", "Link[VarType]"]]:
        return super().__call__(nbunch=nbunch, data=data, default=default)



import pickle
import unittest
from typing import Any

import casadi as cs
import numpy as np
from csnlp import Nlp
from csnlp.wrappers import Mpc
from parameterized import parameterized_class

from sym_metanet import (
    Destination,
    Link,
    LinkWithVsl,
    MeteredOnRamp,
    Network,
    Node,
    engines,
)

with open("tests/data_test_examples.pkl", "rb") as f:
    RESULTS = pickle.load(f)


def create_demands(time: np.ndarray) -> np.ndarray:
    return np.stack(
        (
            np.interp(time, (2.0, 2.25), (3500, 1000)),
            np.interp(time, (0.0, 0.15, 0.35, 0.5), (500, 1500, 1500, 500)),
        )
    )


def get_net(vsl: bool = False) -> tuple[Network, dict[str, Any]]:
    T = 10 / 3600
    Tfin = 2.5
    time = np.arange(0, Tfin, T)
    L = 1
    lanes = 2
    C = (4000, 2000)
    tau = 18 / 3600
    kappa = 40
    eta = 60
    rho_max = 180
    delta = 0.0122
    a = 1.867
    rho_crit = 33.5
    v_free = 102
    N1 = Node(name="N1")
    N2 = Node(name="N2")
    N3 = Node(name="N3")
    O1 = MeteredOnRamp(C[0], name="O1")
    O2 = MeteredOnRamp(C[1], name="O2")
    D1 = Destination(name="D1")
    args = (lanes, L, rho_max, rho_crit, v_free, a)
    if vsl:
        L1 = LinkWithVsl[cs.SX](
            4, *args, segments_with_vsl={2, 3}, alpha=0.1, name="L1"
        )
    else:
        L1 = Link(4, *args, a, name="L1")
    L2 = Link(2, *args, name="L2")
    net = (
        Network(name="A1")
        .add_path(origin=O1, path=(N1, L1, N2, L2, N3), destination=D1)
        .add_origin(O2, N2)
    )
    net.is_valid(raises=True)
    return net, {
        "T": T,
        "L": L,
        "lanes": lanes,
        "C": C,
        "tau": tau,
        "kappa": kappa,
        "eta": eta,
        "rho_max": rho_max,
        "delta": delta,
        "a": a,
        "rho_crit": rho_crit,
        "v_free": v_free,
        "time": time,
    }


@parameterized_class("sym_type", [("SX",), ("MX",)])
class TestExamples(unittest.TestCase):
    def test_dynamics_example(self):
        engine = engines.use("casadi", sym_type=self.sym_type)
        net, pars = get_net()
        net.step(
            T=pars["T"],
            tau=pars["tau"],
            eta=pars["eta"],
            kappa=pars["kappa"],
            delta=pars["delta"],
            engine=engine,
        )
        F = engine.to_function(
            net=net,
            more_out=True,
            compact=1,
            T=pars["T"],
        )
        demands = create_demands(pars["time"]).T
        rho = cs.DM([22, 22, 22.5, 24, 30, 32])
        v = cs.DM([80, 80, 78, 72.5, 66, 62])
        w = cs.DM([0, 0])
        r = cs.DM.ones(2, 1)  # constant ramp metering rates
        RHO, V, W, Q, Q_o = [], [], [], [], []
        for d in demands:
            rho, v, w, q, q_o = F(rho, v, w, r, d)
            RHO.append(rho)
            V.append(v)
            W.append(w)
            Q.append(q)
            Q_o.append(q_o)
        RHO, V, W, Q, Q_o = (np.squeeze(o) for o in (RHO, V, W, Q, Q_o))  # type: ignore
        tts = pars["T"] * sum(
            (rho * pars["L"] * pars["lanes"]).sum() + w.sum() for rho, w in zip(RHO, W)
        )

        RHO_, V_, W_, Q_, Q_o_, tts_ = RESULTS["test_dyn"]
        np.testing.assert_allclose(RHO, RHO_, rtol=1e-3, atol=1e-3)
        np.testing.assert_allclose(V, V_, rtol=1e-3, atol=1e-3)
        np.testing.assert_allclose(W, W_, rtol=1e-3, atol=1e-3)
        np.testing.assert_allclose(Q, Q_, rtol=1e-3, atol=1e-3)
        np.testing.assert_allclose(Q_o, Q_o_, rtol=1e-3, atol=1e-3)
        np.testing.assert_allclose(tts, tts_, rtol=1e-3, atol=1e-3)

    def test_ramp_metering_example(self):
        engine = engines.use("casadi", sym_type=self.sym_type)
        net, pars = get_net()
        net.step(
            T=pars["T"],
            tau=pars["tau"],
            eta=pars["eta"],
            kappa=pars["kappa"],
            delta=pars["delta"],
            engine=engine,
            init_conditions={net.origins_by_name["O1"]: {"r": 1}},
        )
        F = engine.to_function(
            net=net,
            more_out=True,
            compact=2,
            T=pars["T"],
        )
        demands = create_demands(pars["time"]).T

        Np, Nc, M = 7, 3, 6
        mpc = Mpc[cs.SX](
            nlp=Nlp[cs.SX](sym_type=self.sym_type),
            prediction_horizon=Np * M,
            control_horizon=Nc * M,
            input_spacing=M,
        )
        n_seg, n_orig = sum(link.N for _, _, link in net.links), len(net.origins)
        rho, _ = mpc.state("rho", n_seg, lb=0)
        v, _ = mpc.state("v", n_seg, lb=0)
        w, _ = mpc.state(
            "w", n_orig, lb=0, ub=[[np.inf], [100]]
        )  # O2 queue is constrained
        r, _ = mpc.action("r", lb=0, ub=1)
        mpc.disturbance("d", n_orig)
        mpc.set_dynamics(F)
        r_last = mpc.parameter("r_last", (r.size1(), 1))
        mpc.minimize(
            pars["T"] * cs.sum2(cs.sum1(rho * pars["L"] * pars["lanes"]) + cs.sum1(w))
            + 0.4 * cs.sumsqr(cs.diff(cs.horzcat(r_last, r)))
        )
        opts = {
            "expand": True,
            "print_time": False,
            "ipopt": {"max_iter": 500, "sb": "yes", "print_level": 0},
        }
        mpc.init_solver(solver="ipopt", opts=opts)

        rho = cs.DM([22, 22, 22.5, 24, 30, 32])
        v = cs.DM([80, 80, 78, 72.5, 66, 62])
        w = cs.DM([0, 0])
        r_last = cs.DM.ones(r.size1(), 1)
        sol_prev = None
        RHO, V, W, Q, Q_o, R = [], [], [], [], [], []
        for k in range(100):
            d_hat = demands[k : k + Np * M, :]
            if k % M == 0:
                sol = mpc.solve(
                    pars={
                        "rho_0": rho,
                        "v_0": v,
                        "w_0": w,
                        "d": d_hat.T,
                        "r_last": r_last,
                    },
                    vals0=sol_prev,
                )
                sol_prev = sol.vals
                r_last = sol.vals["r"][0]
            x_next, q_all = F(cs.vertcat(rho, v, w), r_last, demands[k, :])
            rho, v, w = cs.vertsplit(x_next, (0, n_seg, 2 * n_seg, 2 * n_seg + n_orig))
            q, q_o = cs.vertsplit(q_all, (0, n_seg, n_seg + n_orig))
            RHO.append(rho)
            V.append(v)
            W.append(w)
            Q.append(q)
            Q_o.append(q_o)
            R.append(r_last)
        RHO, V, W, Q, Q_o, R = (np.squeeze(o) for o in (RHO, V, W, Q, Q_o, R))  # type: ignore
        tts = pars["T"] * sum(
            (rho * pars["L"] * pars["lanes"]).sum() + w.sum() for rho, w in zip(RHO, W)
        )

        RHO_, V_, W_, Q_, Q_o_, R_, tts_ = RESULTS["test_ramp"]
        np.testing.assert_allclose(RHO, RHO_, rtol=1e-3, atol=1e-3)
        np.testing.assert_allclose(V, V_, rtol=1e-3, atol=1e-3)
        np.testing.assert_allclose(W, W_, rtol=1e-3, atol=1e-3)
        np.testing.assert_allclose(Q, Q_, rtol=1e-3, atol=1e-3)
        np.testing.assert_allclose(Q_o, Q_o_, rtol=1e-3, atol=1e-3)
        np.testing.assert_allclose(R, R_, rtol=1e-3, atol=1e-3)
        np.testing.assert_allclose(tts, tts_, rtol=1e-3, atol=1e-3)

    def test_coordinated_control_example(self):
        engine = engines.use("casadi", sym_type=self.sym_type)
        net, pars = get_net(vsl=True)
        net.step(
            T=pars["T"],
            tau=pars["tau"],
            eta=pars["eta"],
            kappa=pars["kappa"],
            delta=pars["delta"],
            engine=engine,
            init_conditions={net.origins_by_name["O1"]: {"r": 1}},
        )
        F = engine.to_function(
            net=net,
            more_out=True,
            compact=2,
            T=pars["T"],
        )
        demands = create_demands(pars["time"]).T

        Np, Nc, M = 7, 5, 6
        mpc = Mpc[cs.SX](
            nlp=Nlp[cs.SX](sym_type=self.sym_type),
            prediction_horizon=Np * M,
            control_horizon=Nc * M,
            input_spacing=M,
        )
        n_seg, n_orig = sum(link.N for _, _, link in net.links), len(net.origins)
        rho, _ = mpc.state("rho", n_seg, lb=0)
        v, _ = mpc.state("v", n_seg, lb=0)
        w, _ = mpc.state(
            "w", n_orig, lb=0, ub=[[np.inf], [100]]
        )  # O2 queue is constrained
        v_ctrl, _ = mpc.action("v_ctrl", 2, lb=20, ub=pars["v_free"])
        r, _ = mpc.action("r", lb=0, ub=1)
        mpc.disturbance("d", n_orig)
        mpc.set_dynamics(F)
        v_ctrl_last = mpc.parameter("v_ctrl_last", (v_ctrl.size1(), 1))
        r_last = mpc.parameter("r_last", (r.size1(), 1))
        mpc.minimize(
            pars["T"] * cs.sum2(cs.sum1(rho * pars["L"] * pars["lanes"]) + cs.sum1(w))
            + 0.4
            * cs.sumsqr(cs.diff(cs.horzcat(v_ctrl_last, v_ctrl), 1, 1) / pars["v_free"])
            + 0.4 * cs.sumsqr(cs.diff(cs.horzcat(r_last, r)))
        )
        opts = {
            "expand": True,
            "print_time": False,
            "ipopt": {"max_iter": 500, "sb": "yes", "print_level": 0},
        }
        mpc.init_solver(solver="ipopt", opts=opts)

        rho = cs.DM([22, 22, 22.5, 24, 30, 32])
        v = cs.DM([80, 80, 78, 72.5, 66, 62])
        w = cs.DM([0, 0])
        v_ctrl_last = v[: net.links_by_name["L1"].N][net.links_by_name["L1"].vsl]
        r_last = cs.DM.ones(r.size1(), 1)
        sol_prev = None
        RHO, V, W, Q, Q_o, V_CTRL, R = [], [], [], [], [], [], []
        for k in range(100):
            d_hat = demands[k : k + Np * M, :]
            if k % M == 0:
                sol = mpc.solve(
                    pars={
                        "rho_0": rho,
                        "v_0": v,
                        "w_0": w,
                        "d": d_hat.T,
                        "v_ctrl_last": v_ctrl_last,
                        "r_last": r_last,
                    },
                    vals0=sol_prev,
                )
                sol_prev = sol.vals
                v_ctrl_last = sol.vals["v_ctrl"][:, 0]
                r_last = sol.vals["r"][0]
            x_next, q_all = F(
                cs.vertcat(rho, v, w), cs.vertcat(v_ctrl_last, r_last), demands[k, :]
            )
            rho, v, w = cs.vertsplit(x_next, (0, n_seg, 2 * n_seg, 2 * n_seg + n_orig))
            q, q_o = cs.vertsplit(q_all, (0, n_seg, n_seg + n_orig))
            RHO.append(rho)
            V.append(v)
            W.append(w)
            Q.append(q)
            Q_o.append(q_o)
            V_CTRL.append(v_ctrl_last)
            R.append(r_last)
        RHO, V, W, Q, Q_o, V_CTRL, R = (np.squeeze(o) for o in (RHO, V, W, Q, Q_o, V_CTRL, R))  # type: ignore
        tts = pars["T"] * sum(
            (rho * pars["L"] * pars["lanes"]).sum() + w.sum() for rho, w in zip(RHO, W)
        )

        RHO_, V_, W_, Q_, Q_o_, V_CTRL_, R_, tts_ = RESULTS["test_coord"]
        np.testing.assert_allclose(RHO, RHO_, rtol=1e-3, atol=1e-3)
        np.testing.assert_allclose(V, V_, rtol=1e-3, atol=1e-3)
        np.testing.assert_allclose(W, W_, rtol=1e-3, atol=1e-3)
        np.testing.assert_allclose(Q, Q_, rtol=1e-3, atol=1e-3)
        np.testing.assert_allclose(Q_o, Q_o_, rtol=1e-3, atol=1e-3)
        np.testing.assert_allclose(R, R_, rtol=1e-3, atol=1e-3)
        np.testing.assert_allclose(V_CTRL, V_CTRL_, rtol=1e-3, atol=1e-3)
        np.testing.assert_allclose(tts, tts_, rtol=1e-3, atol=1e-3)


if __name__ == "__main__":
    unittest.main()
